LIBNAME SGF2016 "/folders/myfolders/SGF2016/collapse";

%macro collapse_data (p_inDsName=, p_outDsName=,p_classVars=, p_prntVar=, p_minLimitCondition=, p_mapDsName=);

	%LOCAL	l_eTime
            l_msg
            l_rc
            l_rTime
			l_sTime
			l_needCollapse
			l_needCollapseCnt 
			l_hashKeyVars
			l_currVarsLenghtStmt
			l_prevVarsLenghtStmt			
			l_currVarsAssignStmt			
			l_prevVarsAssignStmt
			l_currVarsAssignStmt2
			l_prevVarsAssignStmt2
			;

	/********** BEGIN -- Main Macro Processing **********/

	/** Record Starting Time **/
	%let l_sTime = %sysfunc(time());
	
	/* First - Examine we need to Perform the Collapsing Process */
	%let l_needCollapseCnt = 0;
	PROC SQL NOPRINT;
		SELECT STRIP(PUT(COUNT(*),best.))
				,STRIP(PUT(cell_number,BEST.))
		INTO	:l_needCollapseCnt
				,:l_needCollapse separated by  ' '
		FROM	&p_inDsName
		WHERE	&p_minLimitCondition
		;
	QUIT;
	%put ;
	%put >>> --------------------------------------------------------- ;
	%put >>> Cells must be Collapsed Count = &l_needCollapseCnt;
	%PUT >>> Cells must be Collapsed = &l_needCollapse;
	%put >>> --------------------------------------------------------- ;
	%put ;
	
	%if(&l_needCollapseCnt EQ 0) %then
	%do;
		%let l_rc  = 1;
		%let l_msg = ERROR>>> collapse_data : Input table did not contain collapsable records! Process terminated.;
		%goto exit;
	%end;

	/* Initialize Dynamic Variables and Statements to be used later */		
	%let l_libName = WORK;
	%let l_dsName  = %UPCASE(%superq(p_inDsName));
	%if (%INDEX(%superq(l_dsName),%str(.)) GT 0) %then
	%do;
		%let l_libName = %SCAN(%superq(l_dsName),1,%str(.));
		%let l_dsName  = %SCAN(%superq(l_dsName),2,%str(.));
	%end;

	PROC SQL NOPRINT;
		SELECT	CATS('curr_',name)||' '||
				CASE WHEN type = 'char' then '$' ELSE '' END ||
				STRIP(PUT(length,best.))
				,CATS('prev_',name)||' '||
				CASE WHEN type = 'char' then '$' ELSE '' END ||
				STRIP(PUT(length,best.))
				,CATS('curr_',name,' = ',name,';')
				,CATS('prev_',name,' = ',name,';')
				,CATS(name,' = curr_',name,';')
				,CATS(name,' = prev_',name,';')
		INTO	:l_currVarsLenghtStmt separated by ' '
				,:l_prevVarsLenghtStmt separated by ' '
				,:l_currVarsAssignStmt separated by ' '
				,:l_prevVarsAssignStmt separated by ' '
				,:l_currVarsAssignStmt2 separated by ' '
				,:l_prevVarsAssignStmt2 separated by ' '
		FROM	DICTIONARY.COLUMNS
		WHERE	LIBNAME = "&l_libName"
		AND		MEMNAME = "&l_dsName"
		;
	QUIT;

	/* scale_Value unwegt popct wegt origCellNumber */
	
	%let l_hashKeyVars= %str(%')%sysfunc(tranwrd(%superq(p_classVars), %str( ),%str(%',%')))%str(%');

	DATA _NULL_ ;

		/* Define all the variables in the PDV */
		if 0 then SET &p_inDsName;
		
		LENGTH	&l_currVarsLenghtStmt
				&l_prevVarsLenghtStmt 
				usedCellsToCollapseWith $3000 recN 8;
				
		RETAIN	usedCellsToCollapseWith;
		
		FORMAT	scale_Value unwegt popct wegt 16.;

		if (_n_ = 1) then
		do;
	 		dcl hash recs	(dataset:"&p_inDsName",ordered:'a',hashexp: 16) ; /* declare hash, and load the data set into it */
		 	recs.definekey  (%unquote(&l_hashKeyVars)) ; /* key table with KEY variable */
	 		recs.definedata (all:'yes') ;  /* store record data */
	 		recs.definedone () ; /* check validity and instantiate object */
	 		
	 		declare hiter hiRecs("recs");
 		end;
 		
 		call missing(of _all_);
 		
		DO recN=1 by 1 UNTIL(LAST.&p_prntVar);
		
			SET &p_inDsName END=eof;
			BY	&p_classVars;

			/* Load an updated copy of the record */
			rc = recs.find();

			&l_currVarsAssignStmt

			if (MISSING(origCellNumber)) then
				origCellNumber = STRIP(PUT(cell_number,BEST.));

			if (MISSING(curr_origCellNumber)) then
				curr_origCellNumber = STRIP(PUT(curr_cell_number,BEST.));

			if (_N_ GT 1) then
			do;
				/* Reset the Hash Iterator position to current record */
				if (hiRecs.setcur() = 0) then
				do;
					rc = hiRecs.PREV(); /* Load previous record into the PDV */

					&l_prevVarsAssignStmt
					if (MISSING(prev_origCellNumber)) then
						prev_origCellNumber = STRIP(PUT(prev_cell_number,BEST.));

					rc = hiRecs.NEXT(); /* Load current record into the PDV */
				end;
			end;

			/* Check the Collapsing Condition */
			if (&p_minLimitCondition) then 
			do;
				/* Reset the Hash Iterator position to current record */
				if (hiRecs.setcur() = 0) then
				do;
					rc = hiRecs.next(); /* Load next record into the PDV */

					if (MISSING(origCellNumber)) then
						origCellNumber = STRIP(PUT(cell_number,BEST.));
				end;

				/* Find the suitable collapsing cell based on closest scale_value.*/
				if (rc = 0) then
					next_diff = abs(scale_Value - curr_scale_Value);

				prev_diff = abs(curr_scale_Value - prev_scale_Value);

				if ((prev_&p_prntVar = curr_&p_prntVar) AND (prev_diff LT next_diff) OR 
					((prev_diff EQ next_diff) AND (prev_unwegt LT unwegt))
					) then
				do;
					prev_origCellNumber = CATX(',',prev_origCellNumber,curr_origCellNumber);
					prev_scale_value = ((curr_scale_value*curr_wegt)+(prev_scale_value*prev_wegt))/(curr_wegt+prev_wegt);
					prev_unwegt      = SUM(curr_unwegt,prev_unwegt);
					prev_wegt        = SUM(curr_wegt,prev_wegt); 
					prev_popct       = SUM(curr_popct,prev_popct);
					
					&l_prevVarsAssignStmt2
				end;
				else if ((curr_&p_prntVar = &p_prntVar) AND (prev_diff GT next_diff) OR 
					((prev_diff EQ next_diff) AND (prev_unwegt GT unwegt))
					) then
				do;
					origCellNumber = CATX(',',origCellNumber,curr_origCellNumber);
					scale_value    = ((curr_scale_value*curr_wegt)+(scale_value*wegt))/(curr_wegt+wegt);
					unwegt         = SUM(curr_unwegt,unwegt);
					wegt           = SUM(curr_wegt,wegt); 
					popct          = SUM(curr_popct,popct);
				end;

				/* Add/Replace the record in the Hash Object */
				recs.replace();

				/* Need to Remove Currently collapsed record from the Hash Object */
				if ((prev_&p_prntVar = curr_&p_prntVar) OR (curr_&p_prntVar = &p_prntVar)) then
				do;
					&l_currVarsAssignStmt2
					if (recs.find() = 0) then
					do;
						rc= recs.remove();
					end;
				end;

			end;
			else
			do;
				/* Replace the record in the Hash Object */
				recs.replace();
			end;
		END;

		if (eof) then
			recs.output(dataset:"&p_outDsName");
	RUN;

	/* Produce Cell_number & Collapse Index mapping data set */
	DATA &p_mapDsName(keep=index);
		LENGTH index 3;
		SET &p_outDsName(keep=cell_number origCellNumber);

		index = _n_;
		i=1;
		do until(scan(origCellNumber,i,',') eq '');
			OUTPUT;
			i+1;
		end;
	RUN;
	
	%goto finished;

	%exit:
		%put *** collapse_data   ***;
		%put *** l_RC must be zero (0). ***;
		%put *** l_RC= &l_RC. ***;
		%put *** &l_MSG ***;
		%put *** collapse_data   ***;

	%finished:
		/** Record Finish Time **/
		%let l_eTime = %sysfunc(time());
	
		%if (%superq(l_sTime) NE ) %then
		%do;
			/* Calculate Run Time and display it */
			%let l_rTime = %sysfunc(putn(%sysevalf(&l_eTime - &l_sTime),time12.2));
			%put;
			%put >>> ----------------------------------------------- ;
			%put >>> collapse_data :>>> Total RunTime = &l_rTime;
			%put >>> ----------------------------------------------- ;
			%put;
		%end;

%mend collapse_data;

PROC SORT data=sgf2016.input;
	BY cell_number sex age; /* Use these variables for the p_classVars= parameter in the %%collapse_data macro call */
run;

%collapse_data(p_inDsName=sgf2016.input, p_outDsName=work.lev1,
p_classVars=%STR(cell_number sex age), p_prntVar=sex
, p_minLimitCondition=%str(unwegt LT 35 OR (popct/wegt) LT 0.67 OR (popct/wegt) GT 4)
, p_mapDsName=work.map1);

